"""
Profiling statistics processor.

Provides functions to extract and filter performance statistics
from `.prof` files generated by Python's `cProfile`.

Used to retrieve cumulative or total execution times for specific
functions or files based on keyword restrictions.
"""

import io
from pathlib import Path
import pstats


def _process_trace(prof_file):
    """
    Loads and sorts the profile statistics using cumulative time.
    """
    prof_file = str(prof_file.resolve())
    s = io.StringIO()
    sortby = pstats.SortKey.CUMULATIVE
    ps = pstats.Stats(prof_file, stream=s).sort_stats(sortby)
    return ps


def _restrict_trace(prof_file, *restrictions):
    """
    Filters profile entries to only those matching any of the restriction keywords.
    """
    filtered_ps = {}
    ps = _process_trace(prof_file)
    for key, value in ps.stats.items():
        filepath, line, fun = key
        if all(
            restriction in filepath or restriction == fun
            for restriction in restrictions
        ):
            filtered_ps[key] = value
    return filtered_ps


def get_prof_time(prof_file, *restrictions, tottime=False):
    """
    Returns the cumulative (or total) time for a single profile entry
    matching the restrictions.
    """
    if not Path(prof_file).resolve().exists():
        raise FileNotFoundError(f"File {prof_file} does not exist.")

    if len(restrictions) == 0:
        raise NotImplementedError("Please, pass a set of valid restrictions.")

    processed_trace = _restrict_trace(prof_file, *restrictions)

    if len(processed_trace) == 0:
        raise ValueError("No matches found with the given restrictions.")

    if len(processed_trace) > 1:
        raise ValueError("Too many matches found; please refine your restrictions.")

    (_, _, tottime_, cumtime_, _) = list(processed_trace.values())[0]  # type: ignore
    return tottime_ if tottime else cumtime_


def get_commit(benchmark_dir: Path, package: str, full: bool = False):
    git_data_dir = benchmark_dir / "git" / package
    with open(git_data_dir / "HEAD", "r", encoding="utf-8") as rf:
        head = rf.read()

    if head.startswith("ref:"):
        path = head.removeprefix("ref:").strip()
        with open(git_data_dir / path, "r", encoding="utf-8") as rf:
            commit = rf.read()
            return commit if full else commit[:7]

    raise NotImplementedError(f"Unknown HEAD format: {head}")
