"""
Handles loading and querying of stored requests from the file system.

Main functionalities:
- Retrieve one or more requests using normalized identifiers.
- Traverse and list existing request identifiers by month.
- Fetch requests as Python dictionaries or as pandas DataFrames for analysis.
- Supports request variants and subrequests (with partial subrequest support).

Relies heavily on utilities for normalization, validation, and path resolution.
"""

from pathlib import Path
from typing import Dict, Iterator, List, Optional, Union
import warnings
import yaml

import pandas as pd

from opa_tools.config import REQUEST_DIR
from opa_tools.utils import process_dct, DATAFRAME_TYPES
from .transformation import flatten
from .utils import (
    ALLOWED_KINDS,
    normalize_request_identifier,
    format_identifier,
    get_request_location,
    validate_request_kind,
    RequestType,
    RequestNotFoundError,
)

NO_EXIST_ERROR = ("gsv", "opa")


def _yaml_read(path: Path):
    with open(path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)


def _get_subrequests(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    no_exist_ok: bool = False,
) -> Dict:
    raise NotImplementedError(
        "Ability to fetch lists of subrequests not implemented yet. "
        "Please, pass subrequests manually (syntax: YYMM/REQID.SUBREQID)"
    )


def _get_single_request(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    allow_subrequests: bool,
    no_exist_ok: bool,
) -> RequestType:
    request_location = get_request_location(
        name,
        subrequest=subrequest,
        month=month,
        kind=kind,
        allow_subrequests=allow_subrequests,
        no_exist_ok=no_exist_ok,
    )
    # TODO: Checked twice
    if request_location.exists():
        return _yaml_read(request_location) or {}
    if no_exist_ok:
        return {}
    raise RequestNotFoundError(name, subrequest, month, kind)


def _subdir_iterator(path: Path, level=0, only_terminal=False) -> Iterator[Path]:
    if level not in {0, 1}:
        raise NotImplementedError("Level must be either 0 or 1.")
    for subdir in sorted(path.iterdir(), key=lambda p: p.name):
        if not subdir.is_dir():
            continue
        if not only_terminal:
            yield subdir
            continue
        is_terminal_subdir = not any(
            subsubdir.is_dir() for subsubdir in subdir.iterdir()
        )
        if is_terminal_subdir:
            yield subdir
            continue
        elif level > 0:
            yield from _subdir_iterator(subdir, level - 1, only_terminal=True)
            continue


def get_request_identifiers(month: str = "*", subrequests: bool = False):
    if month == "*":
        output = []
        for month in sorted(path.name for path in REQUEST_DIR.iterdir()):
            output.extend(get_request_identifiers(month=month, subrequests=subrequests))
        return output
    month_dir = REQUEST_DIR / month
    if not month_dir.exists():
        raise FileNotFoundError(f"Month directory {month_dir} does not exists.")
    output = []
    for subdir in _subdir_iterator(path=month_dir, level=1, only_terminal=subrequests):
        subdir_parts = subdir.relative_to(REQUEST_DIR).parts
        # TODO: Check that is really a month
        output.append(format_identifier(*subdir_parts))
    return output


def _get_requests(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    allow_subrequests: bool,
) -> List[RequestType]:
    if kind == "*":
        return [
            _get_single_request(
                name=name,
                subrequest=subrequest,
                month=month,
                kind=kind_,
                allow_subrequests=allow_subrequests,
                no_exist_ok=(kind_ not in NO_EXIST_ERROR),
            )
            for kind_ in ALLOWED_KINDS
        ]
    return [
        _get_single_request(
            name=name,
            subrequest=subrequest,
            month=month,
            kind=kind,
            allow_subrequests=allow_subrequests,
            no_exist_ok=False,
        )
    ]


def get_request(
    identifier: Optional[Union[str, int, float]] = None,
    /,
    subrequest: Optional[str] = None,
    month: Optional[str] = None,
    kind: str = "*",
    *,
    allow_subrequests=True,
    flat=False,
    single_as_list=False,  # Has no effect when flat is True
) -> Union[List[RequestType], RequestType]:
    if identifier is None:
        if month is None:
            month = "*"
        output = []
        for identifier in get_request_identifiers(month=month):
            output.append(
                get_request(identifier, kind=kind, flat=flat, single_as_list=True)
            )
        return output
    name, subrequest, month = normalize_request_identifier(
        identifier, subrequest=subrequest, month=month
    )
    validate_request_kind(kind)
    requests = _get_requests(
        name=name,
        subrequest=subrequest,
        month=month,
        kind=kind,
        allow_subrequests=allow_subrequests,
    )
    if flat:
        return flatten(requests)
    if len(requests) == 1 and not single_as_list:
        return requests[0]
    return requests


def get_request_df(month: str = "*") -> pd.DataFrame:
    request_df = pd.DataFrame()
    for identifier in get_request_identifiers(month=month, subrequests=True):
        try:
            request: RequestType = get_request(  # type: ignore
                identifier, flat=True, single_as_list=False
            )
        except RequestNotFoundError:
            warnings.warn(f"Skipped request: {identifier}.")
            continue
        entry = pd.DataFrame.from_dict(  # type: ignore
            process_dct(request, valid_types=DATAFRAME_TYPES, list_wrap=True)
        )
        request_df = pd.concat([request_df, entry], ignore_index=True)
    return request_df
