from collections import defaultdict
from datetime import datetime
from itertools import cycle as i_cycle
import os
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)
from matplotlib.colors import to_rgba


DATAFRAME_TYPES = (str, int, float)


class iterdict(defaultdict):
    def __init__(self, iterable: Iterable, /, cycle=False):
        if cycle:
            iterator = iter(iterable)
        else:
            iterator = i_cycle(iterable)
        self._iterator = iterator
        super().__init__(lambda: next(self._iterator))


class ColorInterpolator:
    """
    A class that generates color gradients for interpolation between white
    and a given target color.

    Attributes:
        min (float): The minimum hue value for interpolation.
        max (float): The maximum hue value for interpolation.
        num (int): The total number of steps for the interpolation.
    """

    def __init__(self, min_: float, max_: float, num_: int, alpha: bool = True):
        """
        Initializes the ColorInterpolator with the specified range and number of steps.

        Args:
            min_ (float): The minimum value of the hue range.
            max_ (float): The maximum value of the hue range.
            num_ (int): The number of steps for the gradient.
        """
        self.min = min_
        self.max = max_
        self.num = num_
        self.alpha = alpha

    def __call__(self, color: str, index: int) -> str:
        """
        Interpolates between white and the specified color based on the index.

        Args:
            color (str): The target color in hexadecimal format (e.g., 'ff0000' for red).
            index (int): The current step in the interpolation (ranging from 0 to num-1).

        Returns:
            str: The interpolated color in hexadecimal format (with alpha).
        """
        # Extract the RGB and alpha components from the provided color
        try:
            r, g, b, a = to_rgba(color)
            r = int(r * 255)
            g = int(g * 255)
            b = int(b * 255)
            a = int(a * 255)
        except ValueError as exc:
            raise ValueError(
                f"Malformed RGB format: {color}. Must be hexadecimal."
            ) from exc

        # Calculate the interpolation factor based on the index
        factor = index / (self.num - 1) if self.num > 1 else 1
        factor = (self.max - self.min) * factor + self.min

        # Adjust the RGB values to create the gradient from white to the target color
        r = int(255 + factor * (r - 255))
        g = int(255 + factor * (g - 255))
        b = int(255 + factor * (b - 255))

        if self.alpha:
            # Return the interpolated color in hex format with alpha channel
            return f"#{r:02x}{g:02x}{b:02x}{a:02x}"

        # Return the interpolated color in hex format without alpha channel
        return f"#{r:02x}{g:02x}{b:02x}"


class EnvironPath(Path):
    _flavour = type(Path())._flavour  # type: ignore # Necessary to avoid TypeError during instantiation
    _environ_var = None
    _initialized = False

    def __new__(cls, environ_var):
        instance = super().__new__(cls)
        instance._initialized = False
        instance._environ_var = environ_var
        return instance

    @staticmethod
    def _get_env_directory(environ_var) -> Path:
        path = os.environ.get(environ_var)
        if path is None:
            raise EnvironmentError(f"Please, set environment variable '{environ_var}'.")
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(
                f"Environment variable {environ_var} set to '{path}', "
                "but directory does not exist."
            )
        return path

    @classmethod
    def _from_parsed_parts(cls, drv, root, parts):
        return Path()._from_parsed_parts(drv, root, parts)  # type: ignore

    def _initialize(self):
        if not self._initialized:
            # Adapted from Path._from_parts
            drv, root, parts = self._parse_args(
                [self._get_env_directory(self._environ_var)]
            )
            self._drv = drv
            self._root = root
            self._parts = parts
            self._initialized = True

    def __getattribute__(self, name: str, /) -> Any:
        _require_initialization = ["_make_child"] + [
            method for method in dir(Path) if not method.startswith("_")
        ]
        if name in _require_initialization and not self._initialized:
            self._initialize()
        return super().__getattribute__(name)

    def __getattr__(self, name: str, /) -> Any:
        if not self._initialized:
            print(name)
            self._initialize()
        super().__getattribute__(name)

    def __repr__(self) -> str:
        return f"<EnvironPath({self._environ_var}, resolved={super().__str__()})>"


def _add_date(datetimes, date):
    if date == "0":
        datetimes.append(datetime(1900, 1, 1))
    elif date == "E":
        datetimes.append(datetime(datetime.now().year + 1, 1, 1))
    else:
        datetimes.append(datetime.fromisoformat(date))


def add_date_parsing_args(parser):
    parser.add_argument(
        "-d",
        "--dates",
        action="store",
        type=str,
        nargs="*",
        metavar="DATES",
    )
    parser.add_argument(
        "-i",
        "--interval",
        action="store_true",
        help="Wether the passed dates are an interval. Then, they must be two and in isoformat.",
    )


def parse_date_interval(dates: Sequence[str]) -> Tuple[datetime, datetime]:
    if len(dates) != 2:
        raise ValueError(
            "When passing an interval, there should only be two dates (min and max)."
        )
    datetimes = []
    try:
        for date in dates:
            _add_date(datetimes, date)
    except ValueError as exc:
        raise ValueError(
            "When passing an interval, dates should be passed in isoformat."
        ) from exc
    return tuple(datetimes)


def process_dct(
    dct: Mapping[Any, Any],
    prefix: str = "",
    valid_types: Optional[Sequence[Type]] = None,
    force_fun: Callable = str,
    list_wrap: bool = False,
    *,
    type_changed_keys: Optional[Sequence] = None,
):
    assert isinstance(dct, dict)
    processed_dict = {}
    for key, value in dct.items():
        processed_key = f"{prefix}{key}"
        if valid_types and force_fun and not isinstance(value, valid_types):
            processed_value = force_fun(value)
            if isinstance(type_changed_keys, set):
                type_changed_keys.add(processed_key)
        else:
            processed_value = value
        if list_wrap:
            processed_value = [processed_value]
        processed_dict[processed_key] = processed_value
    return processed_dict
