from argparse import ArgumentParser
from collections import defaultdict
from pathlib import Path
import typing as T
import os
import yaml

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt

from opa_tools.utils import iterdict, ColorInterpolator

figs = []

COLORS = iterdict(["blue", "red", "green", "orange", "yellow", "purple", "brown"], cycle=False)
MARKERS = iterdict(["s", "o", "^", "v", ">", "<"], cycle=False)

LINES = iterdict(["-", "--", "-.", ":"])
HELPER_FILE = Path(__file__).parent.resolve() / "helperdata.yaml"

DF_COLS = ["method", "ndigests", "init_size", "workers", "weight", "include_io", "benchmark"]

plt.rcParams["font.family"] = "Arial"

DEFAULT_OUTPUT_PATH = Path(__file__).resolve().parent


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "-d",
        "--duplicates-method",
        action="store",
        help="Method to remove duplicates",
        choices=["last", "min", "max", "mean", "meanstd"],
        default="last" ,
    )
    parser.add_argument("--std", action="store_true", help="Include standard dev. in the plot.")
    parser.add_argument("-o", "--output", action="store", default=DEFAULT_OUTPUT_PATH)
    return parser


def get_df() -> DataFrame:
    def read(directory, file):
        with open(directory / f"{file}.yaml", "r", encoding="utf-8") as rf:
            return yaml.safe_load(rf)

    try:
        benchmark_dir = Path(os.environ["BENCHMARK_DIR"]) / "crick"
    except KeyError as exc:
        raise EnvironmentError(
            "Please set environment variable 'BENCHMARK_DIR'."
        ) from exc
    if not benchmark_dir.exists():
        raise FileNotFoundError(
            f"Directory '{benchmark_dir} does not exist."
        )
    data = defaultdict(dict)
    for directory in benchmark_dir.iterdir():
        dname = directory.name
        data[dname]["datetime"] = read(directory, "datetime")
        data[dname]["benchmark"] = read(directory, "benchmark")
        data[dname].update(
            read(directory, "args")
        )
        data[dname].update({
            key: value
            for key, value
            in read(directory, "environ").items()
            if not key.startswith("BASH_")
        })
        data[dname]["argv"] = read(directory, "argv")
    
    # Create dataframe and fill weight with 1 (which was default when producing data)
    df = DataFrame.from_dict(data, orient="index")
    df["weight"] = df["weight"].fillna(1)
    df["include_io"] = df["include_io"].astype(bool).fillna(False)
    
    # Place DF_COLS first
    columns = DF_COLS + [column for column in df.columns if column not in DF_COLS]
    df = df[columns]

    return df


def _group_arrays_by_unique(x, y):
    """
    Given two 1D sequences of the same length:
        x: values that may have repetitions.
        y: corresponding values.
    
    Perform the following operation:
        1. Extract the unique values from x: this will be your new 1D array (let's call it xx_unique).
        2. For each unique value in x, collect all the corresponding entries in y.
        3. Pad the collected y entries with np.nan so that all rows have the same length.
    
    This results in resulting in a 2D array with shape (N, M), where .
        - N is the amount of unique elements in x.
        - M is the maximum repeats of of x.
    
    Returns:
        - np.ndarray (shape (N,)): Unique elements of the x sequence.
        - np.ndarray (shape (M, N)): Grouped positions from y as described above. Contains np.nan values.

    """
    x = np.asarray(x)
    y = np.asarray(y)

    unique_x, inverse_indices = np.unique(x, return_inverse=True)
    counts = np.bincount(inverse_indices)
    dim_M = counts.max()
    dim_N = len(unique_x)

    y_grouped = np.full((dim_M, dim_N), np.nan)

    fill_counts = np.zeros(dim_N, dtype=int)
    for x_val, y_val in zip(x, y):
        idx = np.where(unique_x == x_val)[0][0]
        row = fill_counts[idx]
        y_grouped[row, idx] = y_val
        fill_counts[idx] += 1

    return unique_x, y_grouped


def _select_series_by_method(
    x, y, duplicates_method
) -> T.Tuple[T.Sequence[float], T.Sequence[float], T.Sequence[float] | None]:
    x_unique, y_grouped = _group_arrays_by_unique(x, y)
    y_std = np.nanstd(y_grouped, axis=0)
    if duplicates_method == "last":
        # Take only the last occurrence for xx (and corresponding to yy)
        xx, yy = zip(*sorted(dict(zip(x, y)).items()))
        return xx, yy, y_std
    if duplicates_method == "min":
        return x_unique, np.nanmin(y_grouped, axis=0), y_std
    if duplicates_method == "max":
        return x_unique, np.nanmax(y_grouped, axis=0), y_std
    if duplicates_method == "mean":
        return x_unique, np.nanmean(y_grouped, axis=0), y_std
    if duplicates_method == "meanstd":
        return x_unique, np.nanmean(y_grouped, axis=0), np.nanstd(y_grouped, axis=0)
    raise ValueError(f"Unknown method for duplicates: '{duplicates_method}'.")


def figure1(df, duplicates_method: str = "last"):
    fig, ax = plt.subplots()
    for method, df_method in df.groupby("method"):
        method_color = COLORS[method]
        workers_values = sorted(set(df_method["workers"]))
        interpolator = ColorInterpolator(0.3, 1.0, len(workers_values))
        for workers, df_workers in df_method.groupby("workers"):
            label = (
                f"{method.upper()} (np={workers})"
                if method in ["mp", "mpi"]
                else method.upper()
            )
            df_sorted = df_workers.sort_values("ndigests")
            xx, yy, _ = _select_series_by_method(
                x=df_sorted["ndigests"],
                y=df_sorted["benchmark"],
                duplicates_method=duplicates_method,
            )
            color = interpolator(method_color, workers_values.index(workers))
            ax.plot(xx, yy, marker="o", color=color, label=label)
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.legend(loc="lower right", ncols=2, fontsize=9)
    ax.set_xlabel("Sequence size ($N$)")
    ax.set_ylabel("Sequence update time [s]")
    return fig

def figure2(df, duplicates_method: str = "last"):
    fig, axes = plt.subplots(len(set(df["ndigests"])), figsize=(6, 9), sharex=True)
    for (ndigests, df_ndigests), ax in zip(df.groupby("ndigests"), axes):
        for method, df_method in df_ndigests.groupby("method"):
            color = COLORS[method]
            label = method.upper()
            df_sorted = df_method.sort_values("workers")
            xx, yy, _ = _select_series_by_method(
                x=df_sorted["workers"],
                y=df_sorted["benchmark"],
                duplicates_method=duplicates_method,
            )
            ax.plot(xx, yy, marker="o", color=color, label=label)
            if len(xx) == 1:
                ax.axhline(yy.iloc[0], color=color, ls="--", lw=1)

        ax.set_title(f"$N=10^{int(np.log10(ndigests))}$")
        ax.set_xscale("linear")
        ax.set_yscale("log")
        ax.legend(loc="lower right", ncols=1, fontsize=7)
        ax.set_ylabel("Seq. update time [s]")

    # Set xlabel only for last axis in the loop
    ax.set_xlabel("Number of processes (np)")  # pylint: disable=W0631
    return fig


def figure3(df, weights: T.Sequence[int], duplicates_method: str = "last", include_std: bool = False):
    """As figure2, but with many subplots, one for each k and N."""


    df = df[df["ndigests"] <= 1e7]
    nrow = len(set(df["ndigests"]))
    ncol = len(weights)
    fig, axes = plt.subplots(nrow, ncol, figsize=(4 * ncol, 3 * nrow), sharex=True, sharey=True)

    for j, weight, in enumerate(weights):
        df_weight = df[df["weight"] == weight]
        for i, (ndigests, df_ndigests) in enumerate(df_weight.groupby("ndigests")):
            for method, df_method in df_ndigests.groupby("method"):
                ax = axes[i, j]
                color = COLORS[method]
                label = method.upper()
                df_sorted = df_method.sort_values("workers")
                xx, yy, std = _select_series_by_method(
                    x=df_sorted["workers"],
                    y=df_sorted["benchmark"],
                    duplicates_method=duplicates_method,
                )
                ax.plot(xx, yy, marker="o", color=color, label=label)
                if include_std and std is not None:
                    ax.fill_between(xx, yy, yy + std, color=color, alpha=0.2)
                if len(xx) == 1:
                    ax.axhline(yy[0], color=color, ls="--", lw=1)

            axes[i, j].set_title(f"$N=10^{int(np.log10(ndigests))}$")
            axes[i, j].set_xscale("linear")
            axes[i, j].set_yscale("log")

        axes[0, j].set_title(f"$w={weight}$\n{axes[0, j].get_title()}")
        axes[-1, j].set_xlabel("Number of processes (np)")

    for i in range(nrow):
        axes[i, 0].set_ylabel("Seq. update time [s]")

    fig.legend(*axes[0, 0].get_legend_handles_labels(), ncol=4, loc='lower center', bbox_to_anchor=(0.5, 0.0))
    fig.tight_layout()
    fig.subplots_adjust(bottom=0.05)

    return fig


def main():
    args = _get_parser().parse_args()
    output = Path(args.output)
    duplicates_method = args.duplicates_method
    include_std = args.std

    # Get pandas DataFrame
    df = get_df()

    # Restrict to initial size of 1000 (most of the simulations)
    df = df[df["init_size"] == 1000]

    # Restrict to the simulations that include I/O time (last iteration)
    df = df[df["include_io"]]

    # Sort by simulation time, export to "benchmarks.csv" for easy visualization with Calc
    df.sort_values([
        "weight", "method", "workers", "ndigests", "datetime"
    ]).to_csv("benchmarks.csv")


    df_weight_01 = df[df["weight"] == 1]
    df_weight_24 = df[df["weight"] == 24]

    # Figure with weight 01
    fig1 = figure1(df=df_weight_01, duplicates_method=duplicates_method)
    fig1.suptitle("$w=1$")
    fig1.tight_layout()
    fig1.savefig(output / "Figure1.png", dpi=500)

    # Figure with weight 24
    fig2 = figure1(df=df_weight_24, duplicates_method=duplicates_method)
    fig2.suptitle("$w=24$")
    fig2.tight_layout()
    fig2.savefig(output / "Figure2.png", dpi=500)

    # Figure with all weights
    fig3 = figure3(df, [1, 24], duplicates_method=duplicates_method, include_std=include_std)
    fig3.savefig(output / "Figure3.png", dpi=500)


if __name__ == "__main__":
    main()