from argparse import ArgumentParser
from datetime import date, timedelta
import time
import os

import yaml

from one_pass.opa import Opa
from gsv import GSVRetriever

from opa_tools import benchmark_setup


# Check if we are under an MPI environment
MPI_LAUNCHED = any(
    var in os.environ
    for var in [
        "OMPI_COMM_WORLD_SIZE",  # OpenMPI
        "PMI_SIZE",  # MPICH / IntelMPI
        "MPI_LOCALNRANKS",  # Some others
    ]
)


if MPI_LAUNCHED:
    try:
        from mpi4py import MPI
    except ImportError as exc:
        raise ImportError(
            "Optional dependency mpi4py must be installed if running with mpirun/mpiexec."
        ) from exc
    MPI_COMM = MPI.COMM_WORLD
    MPI_RANK = MPI_COMM.Get_rank()
    MPI_SIZE = MPI_COMM.Get_size()


else:
    MPI = None
    MPI_COMM = None
    MPI_RANK = None
    MPI_SIZE = None

EXPID = "a23s"
STARTDATE = date(1990, 1, 1)

OUTPUT_DIR = "/gpfs/scratch/bsc32/bsc122036/outputs"
os.environ["LD_LIBRARY_PATH"] = (
    f"/gpfs/projects/ehpc01/dte/mars/versions/current/lib64:{os.environ['LD_LIBRARY_PATH']}"
)
os.environ["PATH"] = (
    f"/gpfs/projects/ehpc01/dte/mars/versions/current/bin:{os.environ['PATH']}"
)
os.environ["FDB_HOME"] = f"/gpfs/scratch/ehpc01/experiments/{EXPID}/fdb"

opa_stat = None

REQUESTS = f"""
---

    GSVREQUEST:
      expver: {EXPID}
      activity: highresmip
      resolution: standard
      generation: 2
      realization: 1
      experiment: cont
      model: ifs-nemo
      dataset: climate-dt
      class: d1
      type: fc
      stream: clte
      levtype: hl
      levelist: 100
      date: "{STARTDATE.strftime('%Y%m%d')}"
      time: "0000/to/2300/by/0100"
      param: "ws"
      grid: "0.1/0.1"
      area: "60/-5/50/5"
      method: "nn"
    OPAREQUEST:
      # Hardcoded parameters
      variable: "ws"
      stat: "histogram"
      stat_freq: "weekly"
      output_freq: "weekly"
      time_step: 60 # in minutes
      save: True
      checkpoint: True
      save_filepath: {OUTPUT_DIR}
      checkpoint_filepath: {OUTPUT_DIR}
"""

READ_FROM_DATABRIDGE = False


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("--week", action="store_true", help="Simulate full week")
    parser.add_argument("--write", action="store_true", help="Write NetCDF(s)")
    return parser


def simulate_day(gsvrequest, oparequest, write=False):
    # Get data from gsv
    print("Retrieving data")
    gsv = GSVRetriever()
    data = gsv.request_data(gsvrequest, use_stream_iterator=READ_FROM_DATABRIDGE)

    if write:
        data.to_netcdf(
            os.path.join(
                OUTPUT_DIR, f"{EXPID}_{gsvrequest['date']}_{gsvrequest['param']}.nc"
            )
        )

    print("Computing one-pass")
    # Run One Pass algorithm on a specific stat & variable controlled by the oparequest

    t0 = time.time()

    opa_stat = Opa(oparequest)
    opa_stat.compute(data)

    t1 = time.time()

    return t1 - t0


def simulate_stream(gsvrequest, oparequest, write=False):
    global opa_stat

    # Get data from gsv
    gsv = GSVRetriever()
    startdate = STARTDATE + timedelta(days=((7 - STARTDATE.weekday()) % 7))

    times = []

    for day in range(8):
        gsvrequest_day = gsvrequest.copy()
        day_date = startdate + timedelta(days=day)
        gsvrequest_day["date"] = day_date.strftime("%Y%m%d")

        print(f"Retrieving data for day #{day} ({day_date.strftime('%Y-%m-%d')})")
        data = gsv.request_data(
            gsvrequest_day, use_stream_iterator=READ_FROM_DATABRIDGE
        )

        if write:
            data.to_netcdf(
                os.path.join(
                    OUTPUT_DIR,
                    f"{EXPID}_{day_date.strftime('%Y%m%d')}_{gsvrequest['param']}.nc",
                )
            )

        print("Computing one-pass")
        # Run One Pass algorithm on a specific stat & variable controlled by the oparequest

        t0 = time.time()

        opa_stat = Opa(oparequest)
        opa_stat.compute(data)

        t1 = time.time()

        times.append(t1 - t0)

    return times


def main():
    """Run OPA."""

    args = _get_parser().parse_args()

    # Read opa and gsv requests
    requests = yaml.safe_load(REQUESTS)
    oparequest = requests["OPAREQUEST"]
    gsvrequest = requests["GSVREQUEST"]
    parrequest = {
        "MPI_LAUNCHED": MPI_LAUNCHED,
        "MPI_SIZE": MPI_SIZE,
    }

    if (not MPI_LAUNCHED) or MPI_RANK == 0:
        benchmark_path = benchmark_setup(
            environ=True,
            gsvrequest=gsvrequest,
            oparequest=oparequest,
            parrequest=parrequest,
        )

    if args.week:
        times = simulate_stream(gsvrequest, oparequest, args.write)

    else:
        times = simulate_day(gsvrequest, oparequest, args.write)

    if not isinstance(times, list):
        times = [times]

    if (not MPI_LAUNCHED) or MPI_RANK == 0:
        with open(benchmark_path / "times.txt", "w") as wf:
            wf.write("\n".join(times))


if __name__ == "__main__":
    main()
