import os
import sys

import numpy as np

# Check if we are under an MPI environment
MPI_LAUNCHED = any(var in os.environ for var in [
    "OMPI_COMM_WORLD_SIZE",  # OpenMPI
    "PMI_SIZE",              # MPICH / IntelMPI
    "MPI_LOCALNRANKS",       # Some others
])

if MPI_LAUNCHED:
    from mpi4py import MPI
    MPI_COMM = MPI.COMM_WORLD
    MPI_RANK = MPI_COMM.Get_rank()
    MPI_SIZE = MPI_COMM.Get_size()

else:
    MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None
    MPI_MIN, MPI_MAX = None, None
    MPI_split = lambda _: None  # pylint: disable=C3001

K = 24 * 31 # Number of samples
N = 16 # Number of grid cels
M = 10      # Number of bins

if not MPI_LAUNCHED:
    sys.exit(0)

np.random.seed(10)


chunk_size = N // MPI_SIZE
start_index = MPI_RANK * chunk_size
if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
    end_index = N
else:
    end_index = (MPI_RANK + 1) * chunk_size

l_distribution = np.empty((K, chunk_size if MPI_RANK != MPI_SIZE - 1 else N - start_index))
lf_distribution = np.empty((K * N,))


if MPI_RANK == 0:
    distribution = np.random.normal(size=(K, N))
    print("Distribution 0:\n", distribution)
    np.savetxt(f"distribution.csv", l_distribution, fmt="%.5f", delimiter="\t")
    t_distribution = distribution.T.copy()
else:
    t_distribution = None

# Allocate flat buffer for local chunk
recv_count = end_index - start_index
lt_distribution = np.empty((recv_count, K))

# Scatter rows of the transposed matrix (i.e., columns of the original)
MPI_COMM.Scatter(t_distribution, lt_distribution, root=0)

l_distribution = lt_distribution.T

print(f"Local distribution #{MPI_RANK}:\n", l_distribution)
np.savetxt(f"file_{MPI_RANK}.csv", l_distribution, fmt="%.5f", delimiter="\t")