import ast
from argparse import ArgumentParser, Namespace
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
import numpy as np
from pathlib import Path
from typing import Callable, Optional, Sequence, Tuple
import sys

from matplotlib.pyplot import cm
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.lines as mlines

# Import utility functions for benchmarking from external libraries
from opa_tools.benchmarking.retrieval import get_benchmark_iterator, get_request_df
from opa_tools.benchmarking.processor import get_prof_time
from opa_tools.utils import (
    add_date_parsing_args,
    iterdict,
    parse_date_interval,
)

# Function to initialize the command-line argument parser
def _get_parser():
    parser = ArgumentParser()
    add_date_parsing_args(parser)  # Add date parsing arguments to the parser
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        action="store",
        metavar="OUTPUT",
        default="$HOME/Pictures/world_coverage.png",  # Default output file path
    )
    parser.add_argument(
        "-l", action="count", help="Legend", dest="legend", default=0  # Legend flag (counting occurrences)
    )
    parser.add_argument(
        "--logarithmic", action="store_true", help="Make logarithmic y-axis"  # Logarithmic scale flag
    )
    parser.add_argument(
        "plot_nums",  # Specify plot numbers to generate
        type=int,
        action="store",
        nargs="+",  # Accepts one or more plot numbers
    )
    return parser

# Function to retrieve the benchmark date from a "date.txt" file inside the benchmark directory
def _get_benchmark_date(benchmark_dir):
    with open(benchmark_dir / "date.txt", "r") as rf:
        return rf.read()

# Function to create a color iterator using the Viridis colormap
def _get_color_iterator(cm_name=None, n=10):
    return iter(cm.viridis(np.linspace(0, 1, n)))  # Generates `n` evenly spaced colors

# Function to create legend handles (used for the plot's legend) based on colors, markers, and lines
def _get_handles(colors, markers=None, lines=None):
    handles = []
    if lines is None:
        lines = {}  # Default to an empty dictionary if no line styles are provided
    if markers is None:
        for txt_color, color in colors.items():
            handles.append(mpatches.Patch(color=color, label=txt_color))  # Create patches for colors
    else:
        for txt_color, color in colors.items():
            for txt_ls, marker in markers.items():
                label = str(txt_ls) + " - " + str(txt_color)  # Create a combined label
                label = label.removeprefix(" - ").removesuffix(" - ")
                handles.append(
                    mlines.Line2D(
                        [],
                        [],
                        linestyle=lines.get(txt_ls, "-"),
                        marker=marker,
                        color=color,
                        label=label,
                    )
                )  # Create lines for markers and line styles
    return handles

# Function to include a legend in the plot
def _include_legend(ax, colors, markers=None, lines=None, *, decouple=False, ncols=1, loc="best", _handles=None, **kwargs):
    if decouple:
        handles = []
        handles.extend(
            _get_handles(colors=colors, markers=None, lines=None)
        )
        handles.extend(
            _get_handles(colors={"": "black"}, markers=markers, lines=lines)
        )
    else:
        handles = _get_handles(colors=colors, markers=markers, lines=lines)
    ax.legend(handles=handles, loc=loc, ncols=ncols, **kwargs)  # Adds the legend to the plot

# Function to plot world coverage with a simple scatter plot of coverage vs time
def plot_world_coverage(
    dates: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    interval: Optional[Tuple[datetime, datetime]] = None,
    logarithmic: bool = False,
    legend: bool = False,
    **kwargs,
):
    request_df = get_request_df(dates, exclude, interval)  # Get request DataFrame filtered by date and interval
    num_stats = len(set(request_df["opa_stat"]))  # Number of unique statistics
    colors = iterdict(_get_color_iterator(n=num_stats), cycle=False)  # Generate a color iterator
    print(request_df)
    fig, ax = plt.subplots(figsize=kwargs.get("figsize"))  # Create the figure and axes for plotting
    for benchmark_dir in get_benchmark_iterator(
        dates=dates, exclude=exclude, interval=interval
    ):
        benchmark_date = _get_benchmark_date(benchmark_dir)  # Get benchmark date
        benchmark_name = benchmark_dir.name
        benchmark_req = request_df[
            (request_df["date"] == benchmark_date)
            & (request_df["bname"] == benchmark_name)
        ]
        opa_stat = benchmark_req["opa_stat"].iloc[0]
        gsv_area = ast.literal_eval(benchmark_req["gsv_area"].iloc[0])  # Evaluate the area from string
        world_coverage = (
            (gsv_area[0] - gsv_area[2]) * (gsv_area[3] - gsv_area[1])
        ) / (180 * 360)  # Calculate world coverage
        print(benchmark_date, benchmark_name, opa_stat, gsv_area)
        for file in benchmark_dir.iterdir():
            if file.suffix == ".prof":  # Check if file is a profile file
                time = get_prof_time(file, "one_pass", "compute")  # Get the profile computation time
                color = colors[opa_stat]
                ax.plot([world_coverage], [time], "o", color=color, lw=0)  # Plot the point on the graph
                ax.text(time, 0.5, benchmark_name)  # Add text annotations
                ax.text(time, -0.5, opa_stat)
                print(time)
    if legend:
        _include_legend(ax, colors)  # Add the legend to the plot if enabled
    if logarithmic:
        ax.set_yscale("log")  # Use logarithmic scale on the y-axis if enabled
    return fig  # Return the figure for saving or further manipulation

# Function to plot world coverage with different markers based on checkpoints and saves
def plot_world_coverage_2(
    dates: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    interval: Optional[Tuple[datetime, datetime]] = None,
    logarithmic: bool = False,
    legend: int = 0,
    **kwargs,
):
    request_df = get_request_df(dates, exclude, interval)
    num_stats = len(set(request_df["opa_stat"]))
    colors = iterdict(_get_color_iterator(n=num_stats), cycle=False)
    markers = iterdict(["x", "v", "^"], cycle=False)  # Different markers for each checkpoint/save combination
    print(request_df)
    fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    for benchmark_dir in get_benchmark_iterator(
        dates=dates, exclude=exclude, interval=interval
    ):
        benchmark_date = _get_benchmark_date(benchmark_dir)
        benchmark_name = benchmark_dir.name
        benchmark_req = (
            request_df[
                (request_df["date"] == benchmark_date)
                & (request_df["bname"] == benchmark_name)
            ]
            .iloc[0]
            .to_dict()
        )
        opa_stat = benchmark_req["opa_stat"]
        gsv_area = ast.literal_eval(benchmark_req["gsv_area"])
        world_coverage = (
            (gsv_area[0] - gsv_area[2]) * (gsv_area[3] - gsv_area[1])
        ) / (180 * 360)
        color = colors[opa_stat]
        marker = markers[(
            benchmark_req["opa_checkpoint"],
            benchmark_req["opa_save"],
        )]  # Choose marker based on checkpoint and save
        for file in benchmark_dir.iterdir():
            if file.suffix == ".prof":
                time = get_prof_time(file, "one_pass", "compute")  # Get profile time
                ax.plot([world_coverage], [time], marker=marker, color=color, lw=0, alpha=0.5)  # Plot the point
                ax.text(time, 0.5, benchmark_name)
                ax.text(time, -0.5, opa_stat)
                print(time)
                break
    if legend == 0:
        pass
    elif legend == 1:
        _include_legend(ax, colors)
    elif legend == 2:
        _include_legend(ax, colors, markers, ncols=3, fontsize=8)
    else:
        raise ValueError(
            f"Wrong value for legend: {legend}."
        )
    if logarithmic:
        ax.set_yscale("log")
    return fig

# Dataclass to store the function and name of a plot
@dataclass
class Plot:
    fun: Callable  # The function that generates the plot
    name: str  # The name of the plot

    # Method to generate the plot and save it as a PNG file
    def plot(self, args: Namespace):
        import seaborn as sns
        fun = self.fun
        dates = args.dates
        interval = args.interval
        logarithmic = args.logarithmic
        legend = args.legend
        figpath = Path.home() / "Pictures" / f"{self.name}.png"  # Path to save the figure
        kwargs = {"logarithmic": logarithmic, "legend": legend}
        fig = (
            fun(dates=dates, **kwargs)
            if not interval
            else fun(interval=parse_date_interval, **kwargs)
        )
        sns.set_theme()  # Set the seaborn theme for the plot
        fig.savefig(figpath)  # Save the figure as a PNG file
        print(f"Saved figure: {figpath}")
        return fig

# A dictionary of plots, with a number as the key and a Plot object as the value
PLOTS = {
    1: Plot(plot_world_coverage, "world_coverage"),
    2: Plot(plot_world_coverage_2, "world_coverage_2"),
}

# Main entry point of the program
def main():
    argv = sys.argv[1:]
    args = _get_parser().parse_args(argv)  # Parse the command-line arguments
    plot_nums = args.plot_nums
    for num in plot_nums:
        PLOTS[num].plot(args)  # Generate the plot(s) based on the provided plot number(s)

if __name__ == "__main__":
    main()  # Run the main function when the script is executed
